package wechat

import (
	"encoding/json"
	"fmt"
	"log"
	"time"

	"blog.bing89.com/go/wechat/curl"
)

type Token struct {
	Token     string `json:"access_token"`
	ExpiresIn int64  `json:"expires_in"`
	ExpiredAt int64  `json:"expiredAt"`
	ErrCode   int    `json:"errcode"`
	ErrMsg    string `json:"errmsg"`	
}

type Wechat struct {
	AppID     string
	AppSecret string
	token     Token
}

func NewWechat(id, secret string) (*Wechat, error) {
	w := Wechat{
		AppID:     id,
		AppSecret: secret,
	}
	err := w.checkToken()
	return &w, err
}

func (w *Wechat) checkToken() error {
	if w.token.Token == "" || w.token.ExpiredAt < time.Now().Unix() {
		return w.getToken()
	}
	return nil
}

func (w *Wechat) queryStr() string {
	return fmt.Sprintf("appid=%s&secret=%s", w.AppID, w.AppSecret)
}

func (w *Wechat) queryToken() (string, error) {
	err := w.checkToken()
	return fmt.Sprintf("access_token=%s", w.token.Token), err
}

func (w *Wechat) getToken() error {
	url := fmt.Sprintf("%s?grant_type=client_credential&%s", APIGetAccessToken, w.queryStr())
	data, err := curl.SimpleGet(url)
	if err != nil {
		return err
	}
	token := Token{}
	err = json.Unmarshal(data, &token)
	log.Println("====>token:", string(data), "   err:", err)
	if err == nil {
		w.token = token
		w.token.ExpiredAt = w.token.ExpiresIn + time.Now().Unix()
	}
	return err
}

func (w *Wechat) SendText(to, text string) ([]byte, error) {
	msg := Message{
		ToUser: to,
		Type:   MessageTypeText,
		Text: TextMessage{
			Content: text,
		},
	}
	token, err := w.queryToken()
	if err != nil {
		return nil, err
	}
	url := fmt.Sprintf("%s?%s", APISendMessage, token)
	log.Println("=====>", url)
	msg.Show()
	data, err := curl.Post(url, &msg, nil)
	return data, err
}

func (w *Wechat)GetUserInfo(openID string)(*UserInfo, error){
	token, err := w.queryToken()
	if err != nil {
		return nil, err
	}
	url := fmt.Sprintf("%s?lang=zh_CN&%s&openid=%s", APIGetUserInfo, token, openID)
	data, err := curl.SimpleGet(url)
	if err != nil {
		return nil, err
	}
	ui := UserInfo{}
	err = json.Unmarshal(data, &ui)
	return &ui, err
}

func (w *Wechat)GetGrantToken(code string)(*GrantToken, error){
	url := fmt.Sprintf("%s?%s&code=%s&grant_type=authorization_code", APIGetGrantToken, w.queryStr(), code)

	data, err := curl.SimpleGet(url)
	if err != nil {
		return nil, err
	}
	token := GrantToken{}
	err = json.Unmarshal(data, &token)
	return &token, err
}
//{"expire_seconds":"60","action_name": "QR_STR_SCENE", "action_info": {"scene": {"scene_str": '.$id.'}}}
type Scene struct{
	Str string `json:"scene_str"`
}

type ActionInfo struct {
	ActionInfo Scene `json:"scene"`
}
type qrCodeRequest struct{
	ExpireSeconds int `json:"expire_seconds"`
	Action string `json:"action_name"`
	ActionInfo ActionInfo `json:"action_info"`
}

func (w *Wechat)CreateQRCode()(*Ticket, error){
	req := qrCodeRequest{
		ExpireSeconds: 3600,
		Action: "QR_STR_SCENE",
		ActionInfo: ActionInfo{
			Scene{
				Str: fmt.Sprintf("aaa%d", time.Now().UnixNano()),
			},
		},
	}
	token, err := w.queryToken()
	if err != nil {
		return nil, err
	}
	url := fmt.Sprintf("%s?%s", APICreateQRCode, token)
	data, err := curl.Post(url, req, nil)
	if err != nil {
		return nil, err
	}
	ticket := Ticket{}
	err = json.Unmarshal(data, &ticket)
	return &ticket, err
}